"""
Basic utility functions
"""
from __future__ import annotations

import json

import docker
from cloudfiles import CloudFiles
import cloudvolume as cv


def sendjsonfile(cloudvolume: cv.CloudVolume,
                 filename: str,
                 content: str
                 ) -> None:
    """Stores a json file using CloudFiles.

    Args:
        cloudvolume: A CloudVolume.
        filename: The filename to use under the CloudVolume's cloudpath.
        content: The JSON content of the file as a string.
    """
    cf = CloudFiles(cloudvolume.cloudpath)

    prettycontent = json.loads(content)
    prettycontent = json.dumps(prettycontent,
                               sort_keys=True,
                               indent=2,
                               separators=(',', ': '))
    cf.put(filename, prettycontent,
           cache_control='no-cache',
           content_type='application/json')


def dockerimageID(imagename: str, tag: str) -> str:
    """Fetches the full image ID from a docker image name and tag.

    Requires the current user to have access to the docker daemon.

    Args:
        imagename: The image name to identify.
        tag: The image tag to identify.

    Returns:
        The SHA hash of the docker image under the image name and tag if
        it can be found.

    Raises:
        ValueError: Unable to find the requested image.
    """
    key = f'{imagename}:{tag}'

    client = docker.from_env()

    images = client.images.list()

    imagelookup = dict()
    for image in images:
        for tag in image.tags:
            assert tag not in imagelookup, 'multiple images have the same tag?'
            imagelookup[tag] = image.id

    if key in imagelookup:
        return imagelookup[key]
    else:
        raise ValueError(f'{key} not found in the current docker image list')



"""Storage of a processing step (process) along with its code environments.

Handles storing representations of a processing step that's produced or
modified a CloudVolume. These consist of a process description, code
environment, and a set of parameters.

Typical usage:
    description = 'Properly documenting a process'
    parameters = {'prudence': 10/10}
    code_env = PythonGithubEnv('.')  # path to a github repo directory

    process = Process(description, parameters, code_env)
    logprocess(cloudvolume, process)
"""
from __future__ import annotations

import os
import json
import pkg_resources
from typing import Union
from types import SimpleNamespace
from configparser import ConfigParser

import git
import cloudvolume as cv

from . import utils


__all__ = ['Process', 'PythonGithubEnv', 'DockerEnv',
           'logprocess', 'process_absent']


class CodeEnv:
    """A virtual class for representing a code environment

    Attributes:
        codeptr: Some type of "pointer" to use for determining the code
                 environment.
    """
    def __init__(self, codeptr: str):
        self.codeptr = codeptr

    def log(self) -> tuple[str, str]:
        """Preparing a CodeEnv for logging.

        Logging a code environment consists of specifying a "code environment
        file" that contains all of the required information to recover the
        environment. This method should collect that information from any
        subclass.

        Returns:
            Returns a tuple (filename, contents), where filename identifies
            the filename to store alongside the provenance file with
            detailed information about the code environment, and contents
            is a string that contains the contents of that file.
        """
        return self.filename, self.contents

    @property
    def filename(self):
        """The filename where we should store detailed information"""
        raise NotImplementedError

    @property
    def contents(self):
        """The detailed information to store in a code environment file"""
        raise NotImplementedError


class PythonGithubEnv(CodeEnv):
    """A code environment specified by a python github repository.

    Attributes:
        codeptr: A path to the root github directory of the repository.
        repo: A GitPython representation of the repository.
    """
    def __init__(self, codeptr: str):
        """Initialization.

        Creates the repo attribute from the codeptr argument.

        Arguments:
            codeptr: A path to the root github directory of the repository.
        """
        self.codeptr = codeptr
        self.repo = git.Repo(codeptr)

    @property
    def url(self) -> str:
        """The URL used to access the repository."""
        cfg = self.repo.config_reader()
        return cfg.get('remote "origin"', 'url')

    @property
    def repo_name(self) -> str:
        """The name of the github repository."""
        return repo_name_from_url(self.url)

    @property
    def commithash(self) -> str:
        """The hash of the current commit of the github repo."""
        return self.repo.commit().hexsha

    @property
    def diff(self) -> str:
        """The uncommitted code changes within the current environment."""
        return self.repo.git.diff()

    @property
    def packagelist(self) -> list[tuple[str, str]]:
        """The current environment of python packages"""
        return [(p.project_name, p.version)
                for p in pkg_resources.working_set]

    @property
    def filename(self) -> str:
        return f'{self.repo_name}_{self.commithash}'

    @property
    def contents(self) -> str:
        contents = dict()

        contents['name'] = self.url
        contents['CodeEnv type'] = 'PythonGithub'
        contents['commit hash'] = self.commithash
        contents['diff'] = self.diff
        contents['packages'] = self.packagelist

        return json.dumps(contents)


def repo_name_from_url(repo_url: str) -> str:
    """Extracts the bare repo-name from a URL.

    Args:
        repo_url: The URL used to access a github repository

    Returns:
        A string of the 'base' name of the repository.
    """
    return os.path.basename(repo_url).replace('.git', '')


class DockerEnv(CodeEnv):
    """A code environment specified by a docker image.

    Attributes:
        imagename: The userdefined name of the docker image.
        tag: The user-defined tag to lightly version the docker image.
        imageID: a SHA hash of the image contents.
        include_packages: Whether or not to record the current python
            packages when logging this environment.
    """
    def __init__(self,
                 imagename: str,
                 tag: str,
                 imageID: str,
                 include_packages: bool = False):
        self.imagename = imagename
        self.tag = tag
        self.imageID = imageID
        self.include_packages = include_packages

    @property
    def filename(self) -> str:
        # need to replace '/' with something else to avoid creating extra
        # directories
        return (f'{self.imagename.replace("/", "_")}'
                f'_{self.imageID.replace(":", "")}')

    @property
    def contents(self) -> str:
        contents_dict = {
            'CodeEnv type': 'Docker',
            'image name': self.imagename,
            'tag': self.tag,
            'image ID': self.imageID
            }

        if self.include_packages:
            contents_dict['packages'] = self.packagelist

        return json.dumps(contents_dict)

    @property
    def packagelist(self) -> list[tuple[str, str]]:
        """The environment of python packages."""
        return [(p.project_name, p.version)
                for p in pkg_resources.working_set]


class Process:
    """A representation of a process that affects a CloudVolume.

    Attributes:
        description: A user-defined description of what the process
            accomplishes.
        parameters: The parameters used by the process.
        *code_envs: The code environments used for this process.
    """
    def __init__(self,
                 description: str,
                 parameters: Union[dict, SimpleNamespace, ConfigParser],
                 *code_envs: list[CodeEnv]):
        self.description = description
        self.parameters = parameters
        self.code_envs = code_envs

    def log(self) -> tuple[dict[str, str], list[str]]:
        """Packaging the attributes for logging.

        Returns:
            A dictionary containing the task description, a parameter
            dictionary, and information about the code environment files.
            Also returns a list of strings containing the contents of each
            code environment file.
        """
        params = self.logparams()

        code_envfiles, code_envfilecontents = list(), list()
        for code_env in self.code_envs:
            new_envfile, new_envfilecontents = code_env.log()
            code_envfiles.append(new_envfile)
            code_envfilecontents.append(new_envfilecontents)

        return ({'task': self.description,
                 'parameters': params,
                 'code_envfiles': code_envfiles},
                code_envfilecontents)

    def logparams(self) -> dict:
        """Packaging the parameters for logging.

        Converts different ways of representing parameters into a flat
        dictionary.

        Returns:
            A dictionary containing the parameters.
        """
        if isinstance(self.parameters, dict):
            return self.parameters
        elif type(self.parameters) in [SimpleNamespace, ConfigParser]:
            return vars(self.parameters)
        else:
            raise NotImplementedError('parameter object for process'
                                      f'"{self.description}" has type '
                                      f'{type(self.parameters)},'
                                      ' which is not currently supported')


def logprocess(cloudvolume: cv.CloudVolume,
               process: Process,
               duplicate: bool = False
               ) -> None:
    """Logs a processing step to the provenance file.

    Args:
        cloudvolume: A CloudVolume.
        process: A Process to log.
        duplicate: Whether to log the a process even if it has already been
            logged.
    """
    provenance_dict, envfilecontents = process.log()
    envfilenames = provenance_dict['code_envfiles']

    if duplicate or process_absent(cloudvolume, process):
        logcodefiles(cloudvolume, envfilenames, envfilecontents)
        cloudvolume.provenance.processing.append(provenance_dict)

    else:
        raise AssertionError('duplicate set to False,'
                             f' and process {process.description}'
                             'has already been logged')

    cloudvolume.commit_provenance()


def process_absent(cloudvolume: cv.CloudVolume, process: Process) -> bool:
    """Checks whether a process is contained in a provenance file.

    Parses the provenance file for a given CloudVolume and determines
    whether a process has already been logged. Returns True if the process
    is absent.

    Args:
        cloudvolume: A CloudVolume.
        process: A Process.

    Returns:
        A bool describing whether or not the process is absent from
        the CloudVolume's provenance log.
    """
    logged = cloudvolume.provenance.processing

    def sameproc(loggedprocess: Process):
        return (loggedprocess['task'] == process.description
                and loggedprocess['parameters'] == process.parameters)

    candidates = list(filter(sameproc, logged))

    if len(candidates) == 0:
        return True


def logcodefiles(cloudvolume: cv.CloudVolume,
                 filenames: list[str],
                 filecontents: list[str]
                 ) -> None:
    """Logs the code environment files that haven't been logged already.

    Args:
        cloudvolume: A CloudVolume.
        filenames: A list of filenames to log.
        filecontents: The contents of each file to log.
    """
    absentfilenames, absentfilecontents = list(), list()
    for filename, filecontent in zip(filenames, filecontents):
        if codefile_absent(cloudvolume, filename):
            absentfilenames.append(filename)
            absentfilecontents.append(filecontent)

    logjsonfiles(cloudvolume, absentfilenames, absentfilecontents)


def codefile_absent(cloudvolume: cv.CloudVolume, filename: str) -> bool:
    """Checks whether a code environment has been logged in a CloudVolume.

    Parses the provenance file for a given CloudVolume and determines
    whether a code environment file has already been logged. Returns True if
    the process is absent.

    Args:
        cloudvolume: A CloudVolume.
        filename: A code environment filename.

    Returns:
        A bool describing whether or not the code environment file is absent
        from the CloudVolume's provenance log.
    """
    processes = cloudvolume.provenance.processing
    codefilenames = []
    for process in processes:
        if 'code_envfiles' in process:
            codefilenames.extend(process['code_envfiles'])

    return filename not in codefilenames


def logjsonfiles(cloudvolume: cv.CloudVolume,
                 filenames: list[str],
                 filecontents: list[str]
                 ) -> None:
    """Stores extra JSON files alongside a provenance file.

    Args:
        cloudvolume: A CloudVolume.
        filenames: A list of filenames to log.
        filecontents: The contents of each file to log.
    """
    for filename, filecontent in zip(filenames, filecontents):
        utils.sendjsonfile(cloudvolume, filename, filecontent)



"""Functions for storing useful information in CloudVolume provenance files

It takes significant effort to parse the result of an experiment in a
CloudVolume, and even more effort to remember what was parsed months
ago. Here, we define some simple tools to annotate volumes with a
set of "Notes" that describe the motivation for creating a volume,
the results of an experiment represented by such a volume, or other useful
information. These notes are automatically timestamped to aid interpretation.

Typical usage:

    inference = cv.CloudVolume(cloudpath)
    addmotivation(inference, "testing our segmentation of volume X")

    # Go do a lot of work inspecting the volume

    addresult(inference, "We still see a lot of problems with broken spines")
"""
from __future__ import annotations

import datetime
from enum import Enum
from typing import Optional, Union

import cloudvolume as cv


NOTE_SEP = ' \n '
FIELD_SEP = ';'


__all__ = ['parsenotes', 'note_absent',
           'addmotivation', 'addresult', 'addgeneric']


class NoteType(Enum):
    """An enumeration for classifying note objects.

    This is meant to be useful for visualization that's not implemented yet.
    """
    MOTIVATION = 1
    RESULT = 2
    GENERIC = 3


class Note:
    """A representation of a note added to a provenance file.

    Attributes:
        timestamp: An associated timestamp.
        note_type: A classification of the note's "type".
        content: The written content of the note.
    """
    def __init__(self,
                 timestamp: Union[datetime.datetime, str],
                 note_type: Union[NoteType, int],
                 content: str
                 ):
        """Initializes Notes.

        Attempts to convert the timestamp argument to datetime.datetime,
        and the note_type to a NoteType.

        Args:
            timestamp: An associated timestamp.
            note_type: A classification of the note's "type".
            content: The written content of the note.
        """
        if (timestamp is not None
                and not isinstance(timestamp, datetime.datetime)):
            timestamp = datetime.datetime.fromisoformat(timestamp)

        if not isinstance(note_type, NoteType):
            note_type = NoteType[note_type]

        self.timestamp = timestamp
        self.note_type = note_type
        self.content = content

    def __str__(self):
        """Converts a Note to a continuous string.

        This is used to pack notes within the 'description' field of provenance
        files. Separates the attributes using the field separator defined
        within the notebook module (FIELD_SEP).
        """
        return (f"{str(self.timestamp)}{FIELD_SEP}"
                f"{self.note_type.name}{FIELD_SEP}"
                f"{str(self.content)}")


def parsenotes(cloudvolume: cv.CloudVolume,
               note_sep: str = NOTE_SEP,
               field_sep: str = FIELD_SEP
               ) -> list[Note]:
    """Parses the notes present in a provenance file.

    Creates NoteType.GENERIC notes for text not added by this package (or using
    other separators, etc.).

    Args:
        cloudvolume: A CloudVolume.
        note_sep: The delimeter to use between notes within a provenance
                  file's description field. Defaults to notebook.NOTE_SEP.
        field_sep: The delimeter to use between fields of a note.

    Returns:
        A list of Notes capturing the contents of the current provenance
        description.
    """
    description = cloudvolume.provenance.description

    def hascontent(string: str) -> bool: return len(string) > 0

    possiblenotes = list(filter(hascontent, description.split(note_sep)))

    notes = list()
    for possiblenote in possiblenotes:
        try:
            notes.append(Note(*possiblenote.split(field_sep)))
        except ValueError as e:
            raise(e)
            notes.append(Note(None, NoteType.GENERIC, possiblenote))

    return notes


def addnote(cloudvolume: cv.CloudVolume,
            note_type: NoteType,
            content: str,
            timestamp: Optional[Union[datetime.datetime, str]] = None,
            note_sep: str = NOTE_SEP,
            field_sep: str = FIELD_SEP
            ) -> None:
    """Adds a note to a provenance file.

    This interface is more generic than the other module functions, as it can
    log all kinds of note types.

    Args:
        cloudvolume: A CloudVolume.
        note_type: The type of Note to add.
        content: The content of the note.
        timestamp: The timestamp to associate with the Note.
        note_sep: The delimeter to use between notes within a provenance
                  file's description field. Defaults to notebook.NOTE_SEP.
        field_sep: The delimeter to use between fields of a note.
    """
    timestamp = datetime.datetime.now() if timestamp is None else timestamp
    newnote = Note(timestamp, note_type, content)

    if len(parsenotes(cloudvolume, note_sep, field_sep)) != 0:
        cloudvolume.provenance.description += NOTE_SEP

    cloudvolume.provenance.description += str(newnote)

    cloudvolume.commit_provenance()


def addmotivation(cloudvolume: cv.CloudVolume,
                  content: str,
                  timestamp: Optional[Union[datetime.datetime, str]] = None,
                  note_sep: str = NOTE_SEP,
                  field_sep: str = FIELD_SEP
                  ) -> None:
    """Adds a motivation note to a provenance file.

    Args:
        cloudvolume: A CloudVolume.
        content: The content of the note.
        timestamp: The timestamp to associate with the Note.
        note_sep: The delimeter to use between notes within a provenance
                  file's description field. Defaults to notebook.NOTE_SEP.
        field_sep: The delimeter to use between fields of a note.
    """
    addnote(cloudvolume, NoteType.MOTIVATION, content,
            timestamp, note_sep, field_sep)


def addresult(cloudvolume: cv.CloudVolume,
              content: str,
              timestamp: Optional[Union[datetime.datetime, str]] = None,
              note_sep: str = NOTE_SEP,
              field_sep: str = FIELD_SEP
              ) -> None:
    """Adds a result note to a provenance file.

    Args:
        cloudvolume: A CloudVolume.
        content: The content of the note.
        timestamp: The timestamp to associate with the Note.
        note_sep: The delimeter to use between notes within a provenance
                  file's description field. Defaults to notebook.NOTE_SEP.
        field_sep: The delimeter to use between fields of a note.
    """
    addnote(cloudvolume, NoteType.RESULT, content,
            timestamp, note_sep, field_sep)


def addgeneric(cloudvolume: cv.CloudVolume,
               content: str,
               timestamp: Optional[Union[datetime.datetime, str]] = None,
               note_sep: str = NOTE_SEP,
               field_sep: str = FIELD_SEP
               ) -> None:
    """Adds a generic note to a provenance file.

    Args:
        cloudvolume: A CloudVolume.
        content: The content of the note.
        timestamp: The timestamp to associate with the Note.
        note_sep: The delimeter to use between notes within a provenance
                  file's description field. Defaults to notebook.NOTE_SEP.
        field_sep: The delimeter to use between fields of a note.
    """
    addnote(cloudvolume, NoteType.GENERIC, content,
            timestamp, note_sep, field_sep)


def note_absent(cloudvolume: cv.CloudVolume,
                content: str,
                note_type: Optional[NoteType] = None
                ) -> bool:
    """Checks whether a note is contained in a provenance file.

    Parses the provenance file for a given CloudVolume and determines
    whether a note has already been logged. Returns True if the note
    is absent.

    Args:
        cloudvolume: A CloudVolume.
        content: The content of the Note of interest.
        note_type: The Note's type.

    Returns:
        A bool describing whether or not the note is absent from
        the CloudVolume's provenance log.
    """
    notes = parsenotes(cloudvolume)

    def same_note(note1: Note,
                  content: str,
                  note_type: Optional[NoteType] = None
                  ) -> bool:
        return (note1.content == content
                and (note_type is not None and note1.note_type == note_type))

    filtered = [note for note in notes
                if same_note(note, content, note_type)]

    return len(filtered) == 0



__version__ = '0.1.1'

from . import process
from .process import Process, PythonGithubEnv, DockerEnv
from .process import logprocess, process_absent

from . import notebook
from .notebook import parsenotes, note_absent
from .notebook import addmotivation, addresult, addgeneric

from . import utils
from .utils import dockerimageID



import sys
import setuptools
from setuptools import setup, find_packages


__version__ = "0.1.1"


setup(
    name='provenance-toolbox',
    version=__version__,
    description=('Some basic utilities to standardize the use'
                 ' of CloudVolume provenance files'),
    author='Nicholas Turner',
    author_email='nturner@zetta.ai',
    url='https://github.com/ZettaAI/provenance-toolbox',
    packages=setuptools.find_packages(),
    install_requires=['GitPython>=3.1.26', 'docker>=5.0.3',
                      'cloud-volume']
)
